#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>

#define FILE_NAME L"rally.exe"
#define BUFFER_SIZE 256
#define REG_VALUE_NAME L"HDPath"

typedef long ssize_t;

typedef enum {
	false = 0,
	true = 1
} bool;

const unsigned char xpPatch1AMatch[] = {
		0x55, 0x8B, 0xEC, 0x53, 0x56, 0x57, 0x8B, 0x7D, 0x08, 0xB9, 0x00, 0x03, 0x00, 0x00, 0x66, 0xBA,
		0xC7, 0x03, 0x32, 0xC0, 0xEE, 0x66, 0x42, 0x66, 0x42, 0xFC, 0xEC, 0xAA, 0xEB, 0x00, 0xE2, 0xFA,
		0x5F, 0xFE, 0x5B, 0x5D, 0xC2, 0x04, 0x00
	};

const unsigned char xpPatch1BMatch[] = {
		0x55, 0x8B, 0xEC, 0x57, 0x8B, 0x7D, 0x08, 0xB9, 0x00, 0x03, 0x00, 0x00, 0x66, 0xBA, 0xC7, 0x03,
		0x32, 0xC0, 0xEE, 0x66, 0x42, 0x66, 0x42, 0xFC, 0xEC, 0xAA, 0xEB, 0x00, 0xE2, 0xFA, 0x5F, 0x5D,
		0xC2, 0x04, 0x00, 0x90, 0x90, 0x90, 0x90
	};

const unsigned char xpPatch1W2K[] = {
		0x55, 0x8B, 0xEC, 0x53, 0x56, 0x57, 0x90, 0x90, 0x5F, 0x5E, 0x5B, 0x5D, 0xC2, 0x04, 0x00, 0x90,
		0x43, 0x4D, 0x52, 0x4F, 0x4E, 0x57, 0x32, 0x4B, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
		0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90
	};

const unsigned char xpPatch1Replace[] = {
		0x55, 0x8B, 0xEC, 0x53, 0x56, 0x57, 0x90, 0x90, 0x5F, 0x5E, 0x5B, 0x5D, 0xC2, 0x04, 0x00, 0x90,
		0x43, 0x4D, 0x52, 0x5F, 0x57, 0x49, 0x4E, 0x37, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
		0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90
	};


const unsigned char xpPatch2Match[] = {
		0x85, 0xC0, 0x74, 0x0D, 0x8B, 0x4C, 0x24, 0x10, 0xB8, 0x01, 0x00, 0x00, 0x00, 0x3B, 0xC8, 0x75,
		0x02, 0x33, 0xC0, 0x81, 0xC4, 0x94, 0x00, 0x00, 0x00, 0xC3
	};

const unsigned char xpPatch2W2K[] = {
		0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90,
		0x90, 0x33, 0xC0, 0x81, 0xC4, 0x94, 0x00, 0x00, 0x00, 0xC3
	};

const unsigned char xpPatch2Replace[] = {
		                                                                /*0x6C, 0xA2, 0xFF, 0xFF*/
		0x31, 0xC0, 0x81, 0xC4, 0x94, 0x00, 0x00, 0x00, 0xC3, 0x0F, 0x85, 0x00, 0x00, 0x00, 0x00, 0x8B,
		                                              /*0x5F, 0xA2, 0xFF, 0xFF*/    /*0x11, 0xA1, 0xFF*/
		0x7C, 0x24, 0x14, 0x83, 0xFF, 0x00, 0x0F, 0x84, 0x00, 0x00, 0x00, 0x00, 0xE9, 0x00, 0x00, 0x00,
		/*0x00*/
		0x00
	};
const size_t xpPatch2EntryPoint = 0x09;
const size_t xpPatch2OutJumpAddr1 = 0x0B;
const size_t xpPatch2OutJumpAddr2 = 0x18;
const size_t xpPatch2InJumpAddr = 0x1D;

const unsigned char menuPatchMatch[] = {
		0xBF, 0x08, 0x00, 0x00, 0x00, 0x8B, 0xC8, 0xC1, 0xE1, 0x07
	};

const unsigned char menuPatchReplace[] = {
		0xBF, 0x10, 0x00, 0x00, 0x00, 0x8B, 0xC8, 0xC1, 0xE1, 0x07
	};

const unsigned char palettePatchReplace[] = {
		0xE9, 0x00, 0x00, 0x00, 0x00, 0x90
	};
const size_t palettePatchJumpAddr = 1;

const unsigned char noCDPatch1Match[] = {
		0xE8, 0x92, 0x0D, 0x00, 0x00, 0x85, 0xC0, 0x75
	};

const unsigned char noCDPatch1Replace[] = {
		0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0xEB
	};

const unsigned char noCDPatch2Match[] = {
		0xE8, 0x42, 0x0D, 0x00, 0x00, 0x85, 0xC0, 0x74, 0xB6
	};

const unsigned char noCDPatch2Replace[] = {
		0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90, 0x90
	};

typedef struct {
	enum {
		PSV_INVALID,
		PSV_WIN9X,
		PSV_WINXP,
		PSV_WIN7
	} version;
	long xpPatch1;
	long xpPatch2;
	long menuPatch;
	long palettePatch;
	long noCDPatch1;
	long noCDPatch2;
	unsigned char xpPatch2Data[sizeof(xpPatch2Replace)];
	unsigned char palettePatchData[sizeof(palettePatchReplace)];
} PatchStruct;

int vaswprintf(wchar_t **buf, const wchar_t *format, va_list args) {
	va_list args2 = args;
	int retval;
	
	retval = _vsnwprintf(NULL, 0, format, args2);
	va_end(args2);
	
	*buf = (wchar_t *)malloc((++retval)*sizeof(wchar_t));
	retval = _vsnwprintf(*buf, retval, format, args);
	
	return retval;
}

int aswprintf(wchar_t **buf, const wchar_t *format, ...) {
	va_list args;
	int retval;
	
	va_start(args, format);
	retval = vaswprintf(buf, format, args);
	va_end(args);
	
	return retval;
}

bool fileExists(const wchar_t *path) {
	FILE *f = _wfopen(path, L"rb");
	if (!f)
		return false;
	fclose(f);
	return true;
}

long search_kmp(FILE *f, const unsigned char *pattern, size_t length, ssize_t *pmt) {
	// build the PMT
	pmt[0] = -1;
	pmt[1] = 0;
	{
		size_t i=2, j=0;
		while (i<length)
			if (pattern[i-1] == pattern[j]) {
				pmt[i] = ++j;
				i++;
			} else if (j>0)
				j = pmt[j];
			else {
				pmt[i] = 0;
				i++;
			}
	}
	
	// do the search
	{
		size_t m = ftell(f);
		size_t i = 0;
		bool readNext = true;
		unsigned char s;
		while (true) {
			if (readNext) {
				int ch = fgetc(f);
				if (ch == EOF)
					return EOF;
				s = (unsigned char)ch;
			}
			if (pattern[i] == s)
				if (i == length-1)
					return m;
				else {
					i++;
					readNext = true;
				}
			else {
				m += i - pmt[i]; // m+i == m_ + 2i_ - pmt[i_]
				if (pmt[i] > -1) {
					i = pmt[i]; // m+i == m_ + i_
					readNext = false;
				}
				else { // pmt[i] == 0
					i = 0; // m+i == m_ + i_ + 1
					readNext = true;
				}
			}
		}
	}
}

__inline long search(FILE *f, const unsigned char *pattern, size_t length, bool seekToStart) {
	if (seekToStart)
		fseek(f, 0, SEEK_SET);
	if (length == 0)
		return ftell(f);
	if (length >= 2) {
		ssize_t *pmt = (ssize_t *)malloc(length * sizeof(ssize_t));
		long retval = search_kmp(f, pattern, length, pmt);
		free(pmt);
		return retval;
	}
	// length == 1
	while (true) {
		int ch = fgetc(f);
		if (ch == EOF)
			return EOF;
		if (((unsigned char)ch) == *pattern)
			return (ftell(f)-1);
	}
}

void preparePatch(FILE *f, PatchStruct *retval) {
	static const unsigned char setDisplayMode[] = { 0xFF, 0x55, 0x54 };
	static const unsigned char jnzOpcode[] = { 0x0F, 0x85 };
	
	long jnzDestination;
	long continueDestination;
	
	memcpy(retval->xpPatch2Data, xpPatch2Replace, sizeof(xpPatch2Replace));
	memcpy(retval->palettePatchData, palettePatchReplace, sizeof(palettePatchReplace));
	
	retval->version = PSV_INVALID;
	if (search(f, xpPatch1Replace, sizeof(xpPatch1Replace), true) != EOF) {
		retval->version = PSV_WIN7;
		return;
	}
	
	do {
		retval->xpPatch1 = search(f, xpPatch1AMatch, sizeof(xpPatch1AMatch), true);
		if (retval->xpPatch1 == EOF)
			retval->xpPatch1 = search(f, xpPatch1BMatch, sizeof(xpPatch1BMatch), true);
		if (retval->xpPatch1 != EOF) {
			retval->version = PSV_WIN9X;
			break;
		}
		
		retval->xpPatch1 = search(f, xpPatch1W2K, sizeof(xpPatch1W2K), true);
		if (retval->xpPatch1 != EOF) {
			retval->version = PSV_WINXP;
			break;
		}
	} while (false);
	
	switch (retval->version) {
	case PSV_INVALID:
		return;
	
	case PSV_WIN9X:
		retval->xpPatch2 = search(f, xpPatch2Match, sizeof(xpPatch2Match), true);
		break;
	
	case PSV_WINXP:
		retval->xpPatch2 = search(f, xpPatch2W2K, sizeof(xpPatch2W2K), true);
	}
	if (retval->xpPatch2 == EOF) {
		retval->version = PSV_INVALID;
		return;
	}
	
	retval->menuPatch = search(f, menuPatchMatch, sizeof(menuPatchMatch), true);
	if (retval->menuPatch == EOF) {
		retval->version = PSV_INVALID;
		return;
	}
	
	// look for IDirectDraw::SetDisplayMode() call
	if (search(f, setDisplayMode, sizeof(setDisplayMode), true) == EOF) {
		retval->version = PSV_INVALID;
		return;
	}
	// the file pointer will be after the call, look for first JNZ
	retval->palettePatch = search(f, jnzOpcode, sizeof(jnzOpcode), false);
	continueDestination = retval->palettePatch + sizeof(palettePatchReplace);
	// the file pointer will be at the operand
	{
		INT32 jnzDestTmp;
		fread(&jnzDestTmp, 4, 1, f);
		jnzDestination = continueDestination + jnzDestTmp;
	}
	
	// fill the xpPatch2
	*(INT32 *)(retval->xpPatch2Data + xpPatch2OutJumpAddr1) = jnzDestination - (retval->xpPatch2 + xpPatch2OutJumpAddr1 + 4);
	*(INT32 *)(retval->xpPatch2Data + xpPatch2OutJumpAddr2) = jnzDestination - (retval->xpPatch2 + xpPatch2OutJumpAddr2 + 4);
	*(INT32 *)(retval->xpPatch2Data + xpPatch2InJumpAddr) = continueDestination - (retval->xpPatch2 + xpPatch2InJumpAddr + 4);
	
	// fill the palettePatch
	*(INT32 *)(retval->palettePatchData + palettePatchJumpAddr) = retval->xpPatch2 + xpPatch2EntryPoint - (retval->palettePatch + palettePatchJumpAddr + 4);
	
	retval->noCDPatch1 = search(f, noCDPatch1Match, sizeof(noCDPatch1Match), true);
	retval->noCDPatch2 = search(f, noCDPatch2Match, sizeof(noCDPatch2Match), true);
}

bool applyPatch(FILE *f, const PatchStruct *patch) {
	if (patch->version != PSV_WIN7) {
		fseek(f, patch->xpPatch1, SEEK_SET);
		if (fwrite(xpPatch1Replace, sizeof(xpPatch1Replace), 1, f) != 1)
			return false;
		
		fseek(f, patch->xpPatch2, SEEK_SET);
		if (fwrite(patch->xpPatch2Data, sizeof(xpPatch2Replace), 1, f) != 1)
			return false;
		
		fseek(f, patch->menuPatch, SEEK_SET);
		if (fwrite(menuPatchReplace, sizeof(menuPatchReplace), 1, f) != 1)
			return false;
		
		fseek(f, patch->palettePatch, SEEK_SET);
		if (fwrite(patch->palettePatchData, sizeof(palettePatchReplace), 1, f) != 1)
			return false;
	}
	
	if (patch->noCDPatch1 != EOF && patch->noCDPatch2 != EOF) {
		fseek(f, patch->noCDPatch1, SEEK_SET);
		if (fwrite(noCDPatch1Replace, sizeof(noCDPatch1Replace), 1, f) != 1)
			return false;

		fseek(f, patch->noCDPatch2, SEEK_SET);
		if (fwrite(noCDPatch2Replace, sizeof(noCDPatch2Replace), 1, f) != 1)
			return false;
	}
	
	return true;
}

int wWinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PWSTR lpCmdLine, int nCmdShow) {
	wchar_t *filePath = NULL;
	if (fileExists(FILE_NAME))
		filePath = FILE_NAME;
	else {
		HKEY hKey;
		if (RegOpenKeyExW(HKEY_LOCAL_MACHINE, L"Software\\Codemasters\\Rally\\1.00", 0, KEY_READ, &hKey) == ERROR_SUCCESS) {
			DWORD tmpSize = 0;
			DWORD type = REG_SZ;
			do {
				if (RegQueryValueExW(hKey, REG_VALUE_NAME, NULL, &type, NULL, &tmpSize) != ERROR_SUCCESS)
					break;
				filePath = (wchar_t *)malloc(tmpSize);
				
				if (RegQueryValueExW(hKey, REG_VALUE_NAME, NULL, &type, (LPBYTE)filePath, &tmpSize) == ERROR_SUCCESS) {
					size_t len1 = wcslen(filePath);
					size_t len2 = wcslen(FILE_NAME);
					filePath = realloc(filePath, (len1 + len2 + 2)*sizeof(wchar_t));
					filePath[len1] = L'\\';
					wcscpy(filePath + len1 + 1, FILE_NAME);
				} else {
					free(filePath);
					filePath = NULL;
				}
			} while (false);
			RegCloseKey(hKey);
		}
	}
	
	if (!filePath) {
		MessageBoxW(NULL,
				L"Colin McRae Rally executable not found at neither current directory nor install location.\n"
				L"Please ensure that the game is installed or run the patcher in the directory containing '" FILE_NAME L"' file.",
				NULL, MB_OK | MB_ICONERROR);
		return -1;
	}
	
	{
		wchar_t *msg = NULL;
		aswprintf(&msg, L"Colin McRae Rally executable found: %s\nProceed with patching?", filePath);
		if (MessageBoxW(NULL, msg, L"Confirmation", MB_YESNO) != IDYES)
			return 0;
		free(msg);
	}
	
	{
		bool patchOK = false;
		bool alreadyPatched = false;
		PatchStruct patch;
		FILE *f = _wfopen(filePath, L"r+b");
		long pos;
		
		if (!f) {
			MessageBoxW(NULL, L"Could not open the file for writing.", NULL, MB_OK | MB_ICONERROR);
			return -2;
		}
		
		preparePatch(f, &patch);
		if (patch.version != PSV_INVALID) {
			if (patch.version == PSV_WIN7)
				alreadyPatched = true;
			if (patch.noCDPatch1 != EOF && patch.noCDPatch2 != EOF)
				if (MessageBoxW(NULL, L"Would you also like to apply the NoCD patch?", L"NoCD patch", MB_YESNO) == IDYES)
					alreadyPatched = false;
				else {
					patch.noCDPatch1 = EOF;
					patch.noCDPatch2 = EOF;
				}
			if (!alreadyPatched)
				patchOK = applyPatch(f, &patch);
		}
		else if (patch.version != PSV_INVALID)
			patchOK = applyPatch(f, &patch);
		fclose(f);
		
		if (patchOK) {
			MessageBoxW(NULL, L"Patch applied successfully.", L"Success", MB_OK | MB_ICONINFORMATION);
			return 0;
		}
		
		if (alreadyPatched) {
			MessageBoxW(NULL, L"The file is already patched.", L"Information", MB_OK | MB_ICONINFORMATION);
			return 0;
		}
	}
	
	MessageBoxW(NULL, L"Could not patch the file.", NULL, MB_OK | MB_ICONERROR);
	return -3;
}